/*----------------------------------------------------------------------------------------

	Introduction to TigerSHARC Multiprocessing Systems Using VisualDSP++
	Maikel K.B., DSP-Europe 
	June 2002

	ID1.C

	This program demonstrates some of the Multiprocessing features handled by VisualDSP++ Tools.
	It shows the use of TigerSHARC in an MP System by performing the following tasks: 

		*	Setting up of control registers: SYSCON and SDRCON
	
		*	ID Checking routine to ensure that the code has been loaded into the right DSP
		
		*	It performs a Vector Interrupt (VIRPT) in DSP ID0 by writing the address of the 
			Interrupt Service Routine (ISR) to be executed in ID0 to the VIRPT register also
			in ID0. 

		*	It copies the contents of tx_ID1 to	the shared external	memory (shared_data 
			buffer declared in SDRAM) using DMA transfer. 
			
		*	Also, it performs a core write from tx_ID1 to rx_ID0 (internal memory of DSP ID1 
			via the MMS) using the Integer ALU (IALU) and circular buffers.
 	
		* 	After completion, it goes into an endless loop.
		
------------------------------------------------------------------------------------------*/

#define __NOUNDERSCORES__							// Depending on the compiler patch installed, the register names as defined in sysreg.h 
													// may now have two underscores prepended so that they do not restrict the namespace 
													// of user programs. If the old names are required as in this example(without the 
													// underscores prepended), 
													// define the preprocessor macro __NOUNDERSCORES__ before including sysreg.h.
													// Only required with compiler patch 6.1.7 or later

#include "sysreg.h"
#include "signal.h"
#include "stdio.h"
#include "defts101.h"

#define REG(ADDR)	(*(int*)(ADDR))					// This allows to directly write to the VIRPT_LOC address defined in defts101.h
#define TAPS 16										// Number of words to transfer
#define MMS_ID0 0x2000000							// ID0 Multiprocessor Memory Space (MMS) offset value	

extern int shared_data[TAPS*2];						// Data array defined in external shared memory (SDRAM)
extern int rx_ID0[TAPS]; 							// ID0 receive buffer (ID0.c) for core (IALU) data transfer from ID1

/* DM data section in Memory Block 1 */
section ("data1") int dm tx_ID1[TAPS] = {			// ID1 transmit buffer for external port (DMA) data 
#include "tx_ID1.dat"								// transfer to external shared memory (second half of buffer)
};													// and CORE transfer to ID0
section ("data1") int (*Ptr_to_Func)();				// it declares a pointer to function 

/* DM data section in Memory Block 2 */
section ("data2") int dm Broadcast_rx_ID1[TAPS];	// ID1 broadcast receive buffer for broadcast transfer from ID0
section ("data2") int dm rx_ID1[TAPS];				// ID1 receive buffer for DMA transfer from ID0


struct TCB {
	int *DI;										// index
	int DX;											// count and stride in x direction
	int DY;											// count and stride in y direction
	int DP;											// DMA control word
};

struct TCB TCB_temp;								// Temp structure for programming TCBs

__builtin_quad q;									// Temp quad for programming TCBs

void Check_ID_Number();								// Prototype for Check ID Number routine
void ID1_to_SDRAM_DMA();							// Prototype for ID1 to SDRAM DMA
void ID1_to_ID0_CORE();								// Prototype for ID1 to ID0 CORE transfer
void dma0_int();									// Prototype for DMA0 interrupt service routine

void main(void)
{
	int i=0, IMASKL_temp=0;	

	__builtin_sysreg_write(SYSCON,0x003a79e7);		// Initial value after reset (0x279e7) modified 
													// for 64-bit Bus width configuration
								
	__builtin_sysreg_write(SDRCON,0x00005913);		// Enable SDRAM: ENA=1, CAS=2 CL, pipedepth=0,			
										 			// page=512w, ref.rate=1200, trp=2, tras=5, init=1

	/* Enable DMA0 and DMA1 interrupts */	

	IMASKL_temp = __builtin_sysreg_read(IMASKL);	// Read the value of IMASKL
	IMASKL_temp = IMASKL_temp | 0x0000C000;			// Enable DMA0 & DMA1
	__builtin_sysreg_write(IMASKL,IMASKL_temp);		// Write value back to IMASKL

	Check_ID_Number();								// Check Code loaded into right DSP

	do												// Wait until Ptr_to_Func is loaded with the address 
	{	i = (int)Ptr_to_Func;}						// of the VIRPT ISR in ID0. That is, until
	while(i==0);									// "Ptr_to_Func = &VIRPT_ISR_ID0;" in ID0.c gets executed

	REG(VIRPT_LOC+MMS_ID0) = (int)Ptr_to_Func;		// Writes address of the ISR to be serviced by ID0 (VIRPT)

	ID1_to_SDRAM_DMA();								// Perform DMA transfer from ID1 to SDRAM
	ID1_to_ID0_CORE();								// Perform CORE transfer from ID1 to ID0

	while(1)										// Enter Endless loop
		i=i-i;
}

/*********************************************************************************************************************************************/
/*	ID Checking Routine ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void Check_ID_Number()
{		
	int IDx=0;
	int SYSTAT_ID_bits=0x7;											// ID value in SYSTAT bits 2-0 -> 0111 = 0x7
	int i=0;
	
	IDx = (__builtin_sysreg_read(SYSTAT) & SYSTAT_ID_bits);			// read SYSTAT value and extract ID number 

	if (IDx != 1)													// is this DSP with ID1? 
	{	
		printf("\n ID1 code - Executable loaded into wrong DSP \n"); 
		while(1)
			i=i-i;
	}

	return;
}

/*********************************************************************************************************************************************/
/*	ID1 to SDRAM DMA0 Transfer ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void ID1_to_SDRAM_DMA()				
{		
	interrupt(SIGDMA0, dma0_int);					// Assign isr to DMA channel 0
													
	TCB_temp.DI = tx_ID1;							// index points to source buffer in ID1
	TCB_temp.DX = 4 | (TAPS << 16);					// modify is 4 for quad-word transfers, count is TAPS and must be shifted to upper half 
	TCB_temp.DY = 0;								// only a 1 dimension DMA
	TCB_temp.DP = 0x47000000;						// control word set for quad-word transfers to internal memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCS0, q);				// program the TCBs

	TCB_temp.DI = shared_data+TAPS;					// index points to second half of buffer in SDRAM
	TCB_temp.DP = 0x87000000;						// control word set for quad-word transfers to external memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCD0, q);				// program the TCBs

	return;
}

/*********************************************************************************************************************************************/
/*	ID1 to IDO CORE Transfer ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void ID1_to_ID0_CORE()					
{		
	int i=0;

	for (i = 0; i < TAPS; i++)
	{
      rx_ID0[i] = tx_ID1[i];		// Core data transfer from ID1 to ID0
	}
	
	printf("\nExternal memory CORE transfer from ID1 to ID0 completed\n");    	
	return;
}

/*********************************************************************************************************************************************/
/*	DMA0 Interrupt Service Routine ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void dma0_int()						
{		
	printf("\nExternal memory DMA from ID1 to SDRAM completed\n"); 
	return;
}




